countries_dict = {
    "AF": "Afganistán",
    "AL": "Albania",
    "DZ": "Argelia",
    "AS": "Samoa Americana",
    "AD": "Andorra",
    "AO": "Angola",
    "AI": "Anguila",
    "AQ": "Antártida",
    "AG": "Antigua y Barbuda",
    "AR": "Argentina",
    "AM": "Armenia",
    "AW": "Aruba",
    "AU": "Australia",
    "AT": "Austria",
    "AZ": "Azerbaiyán",
    "BS": "Bahamas",
    "BH": "Baréin",
    "BD": "Bangladés",
    "BB": "Barbados",
    "BY": "Bielorrusia",
    "BE": "Bélgica",
    "BZ": "Belice",
    "BJ": "Benín",
    "BM": "Bermudas",
    "BT": "Bután",
    "BO": "Bolivia",
    "BA": "Bosnia y Herzegovina",
    "BW": "Botsuana",
    "BV": "Isla Bouvet",
    "BR": "Brasil",
    "IO": "Territorio Británico del Océano Índico",
    "BN": "Brunéi",
    "BG": "Bulgaria",
    "BF": "Burkina Faso",
    "BI": "Burundi",
    "CV": "Cabo Verde",
    "KH": "Camboya",
    "CM": "Camerún",
    "CA": "Canadá",
    "KY": "Islas Caimán",
    "CF": "República Centroafricana",
    "TD": "Chad",
    "CL": "Chile",
    "CN": "China",
    "CX": "Isla de Navidad",
    "CC": "Islas Cocos",
    "CO": "Colombia",
    "KM": "Comoras",
    "CG": "República del Congo",
    "CD": "República Democrática del Congo",
    "CK": "Islas Cook",
    "CR": "Costa Rica",
    "CI": "Costa de Marfil",
    "HR": "Croacia",
    "CU": "Cuba",
    "CY": "Chipre",
    "CZ": "Chequia",
    "DK": "Dinamarca",
    "DJ": "Yibuti",
    "DM": "Dominica",
    "DO": "República Dominicana",
    "EC": "Ecuador",
    "EG": "Egipto",
    "SV": "El Salvador",
    "GQ": "Guinea Ecuatorial",
    "ER": "Eritrea",
    "EE": "Estonia",
    "ET": "Etiopía",
    "FK": "Islas Malvinas",
    "FO": "Islas Feroe",
    "FJ": "Fiyi",
    "FI": "Finlandia",
    "FR": "Francia",
    "GF": "Guayana Francesa",
    "PF": "Polinesia Francesa",
    "TF": "Territorios Australes Franceses",
    "GA": "Gabón",
    "GM": "Gambia",
    "GE": "Georgia",
    "DE": "Alemania",
    "GH": "Ghana",
    "GI": "Gibraltar",
    "GR": "Grecia",
    "GL": "Groenlandia",
    "GD": "Granada",
    "GP": "Guadalupe",
    "GU": "Guam",
    "GT": "Guatemala",
    "GG": "Guernesey",
    "GN": "Guinea",
    "GW": "Guinea-Bisáu",
    "GY": "Guyana",
    "HT": "Haití",
    "HM": "Islas Heard y McDonald",
    "HN": "Honduras",
    "HK": "Hong Kong",
    "HU": "Hungría",
    "IS": "Islandia",
    "IN": "India",
    "ID": "Indonesia",
    "IR": "Irán",
    "IQ": "Irak",
    "IE": "Irlanda",
    "IM": "Isla de Man",
    "IL": "Israel",
    "IT": "Italia",
    "JM": "Jamaica",
    "JP": "Japón",
    "JE": "Jersey",
    "JO": "Jordania",
    "KZ": "Kazajistán",
    "KE": "Kenia",
    "KI": "Kiribati",
    "KP": "Corea del Norte",
    "KR": "Corea del Sur",
    "KW": "Kuwait",
    "KG": "Kirguistán",
    "LA": "Laos",
    "LV": "Letonia",
    "LB": "Líbano",
    "LS": "Lesoto",
    "LR": "Liberia",
    "LY": "Libia",
    "LI": "Liechtenstein",
    "LT": "Lituania",
    "LU": "Luxemburgo",
    "MO": "Macao",
    "MG": "Madagascar",
    "MW": "Malaui",
    "MY": "Malasia",
    "MV": "Maldivas",
    "ML": "Malí",
    "MT": "Malta",
    "MH": "Islas Marshall",
    "MQ": "Martinica",
    "MR": "Mauritania",
    "MU": "Mauricio",
    "YT": "Mayotte",
    "MX": "México",
    "FM": "Micronesia",
    "MD": "Moldavia",
    "MC": "Mónaco",
    "MN": "Mongolia",
    "ME": "Montenegro",
    "MS": "Montserrat",
    "MA": "Marruecos",
    "MZ": "Mozambique",
    "MM": "Birmania",
    "NA": "Namibia",
    "NR": "Nauru",
    "NP": "Nepal",
    "NL": "Países Bajos",
    "NC": "Nueva Caledonia",
    "NZ": "Nueva Zelanda",
    "NI": "Nicaragua",
    "NE": "Níger",
    "NG": "Nigeria",
    "NU": "Niue",
    "NF": "Isla Norfolk",
    "MK": "Macedonia del Norte",
    "MP": "Islas Marianas del Norte",
    "NO": "Noruega",
    "OM": "Omán",
    "PK": "Pakistán",
    "PW": "Palaos",
    "PS": "Palestina",
    "PA": "Panamá",
    "PG": "Papúa Nueva Guinea",
    "PY": "Paraguay",
    "PE": "Perú",
    "PH": "Filipinas",
    "PN": "Islas Pitcairn",
    "PL": "Polonia",
    "PT": "Portugal",
    "PR": "Puerto Rico",
    "QA": "Catar",
    "RE": "Reunión",
    "RO": "Rumanía",
    "RU": "Rusia",
    "RW": "Ruanda",
    "BL": "San Bartolomé",
    "SH": "Santa Elena",
    "KN": "San Cristóbal y Nieves",
    "LC": "Santa Lucía",
    "MF": "San Martín (Francia)",
    "PM": "San Pedro y Miquelón",
    "VC": "San Vicente y las Granadinas",
    "WS": "Samoa",
    "SM": "San Marino",
    "ST": "Santo Tomé y Príncipe",
    "SA": "Arabia Saudita",
    "SN": "Senegal",
    "RS": "Serbia",
    "SC": "Seychelles",
    "SL": "Sierra Leona",
    "SG": "Singapur",
    "SX": "San Martín (Países Bajos)",
    "SK": "Eslovaquia",
    "SI": "Eslovenia",
    "SB": "Islas Salomón",
    "SO": "Somalia",
    "ZA": "Sudáfrica",
    "GS": "Islas Georgia del Sur y Sandwich del Sur",
    "SS": "Sudán del Sur",
    "ES": "España",
    "LK": "Sri Lanka",
    "SD": "Sudán",
    "SR": "Surinam",
    "SJ": "Svalbard y Jan Mayen",
    "SE": "Suecia",
    "CH": "Suiza",
    "SY": "Siria",
    "TW": "Taiwán",
    "TJ": "Tayikistán",
    "TZ": "Tanzania",
    "TH": "Tailandia",
    "TL": "Timor Oriental",
    "TG": "Togo",
    "TK": "Tokelau",
    "TO": "Tonga",
    "TT": "Trinidad y Tobago",
    "TN": "Túnez",
    "TR": "Turquía",
    "TM": "Turkmenistán",
    "TC": "Islas Turcas y Caicos",
    "TV": "Tuvalu",
    "UG": "Uganda",
    "UA": "Ucrania",
    "AE": "Emiratos Árabes Unidos",
    "GB": "Reino Unido",
    "US": "Estados Unidos",
    "UM": "Islas Menores Alejadas de los Estados Unidos",
    "UY": "Uruguay",
    "UZ": "Uzbekistán",
    "VU": "Vanuatu",
    "VE": "Venezuela",
    "VN": "Vietnam",
    "VG": "Islas Vírgenes Británicas",
    "VI": "Islas Vírgenes de los Estados Unidos",
    "WF": "Wallis y Futuna",
    "EH": "Sahara Occidental",
    "YE": "Yemen",
    "ZM": "Zambia",
    "ZW": "Zimbabue",
    "XK": "Kosovo",
    "CW": "Curazao",
    "VA": "Ciudad del Vaticano",
    "SZ": "Esuatini",
}



from datetime import datetime
from typing import Generator

from domain.models.base_model import QueryParams, Affiliation, Author, ExternalId
from domain.models.work_model import Work, Source
from infrastructure.repositories import api_expert_repository
from domain.parsers import work_parser


def get_works_by_person(person_id: str, query_params: QueryParams) -> dict:
    works = api_expert_repository.get_works_by_person_for_api_expert(person_id, query_params)
    data = process_works(works)
    return {"data": data}


def get_works_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    works = api_expert_repository.get_works_by_affiliation_for_api_expert(affiliation_id, query_params)
    data = process_works(works)
    return {"data": data}


def search_works(query_params: QueryParams) -> dict:
    works = api_expert_repository.search_works_for_api_expert(query_params)
    data = process_works(works)
    return {"data": data}


def process_works(works: Generator) -> list:
    works_list = []
    for work in works:
        set_authors_data(work)
        set_authors_affiliations_data(work)
        set_source_data(work)
        works_list.append(work)
    data = work_parser.parse_api_expert(works_list)
    return data


def set_authors_data(work: Work) -> None:
    for author in work.authors:
        author_data = Author()
        if work.authors_data:
            author_data = next(filter(lambda x: x.id == author.id, work.authors_data), Author())
        author.external_ids = author_data.external_ids
        author.ranking = author_data.ranking
        author.last_names = author_data.last_names
        author.first_names = author_data.first_names
        author.sex = author_data.sex
        author.affiliations = author_data.affiliations
        if author_data.birthplace:
            author.birth_country = author_data.birthplace.country
        if author_data.birthdate and author_data.birthdate != -1 and author_data.birthdate != "":
            birthdate = datetime.fromtimestamp(author_data.birthdate)
            today = datetime.today()
            age = today.year - birthdate.year
            if (today.month, today.day) < (birthdate.month, birthdate.day):
                age -= 1
            author.age = age
    work.authors_data = None


def set_authors_affiliations_data(work: Work) -> None:
    for author in work.authors:
        author.countries = []
        for affiliation in author.affiliations:
            affiliation_data = Affiliation()
            if work.affiliations_data:
                affiliation_data = next(filter(lambda x: x.id == affiliation.id, work.affiliations_data), Affiliation())
            if affiliation_data.external_ids:
                affiliation.ror = next(
                    filter(lambda x: x.source == "ror", affiliation_data.external_ids), ExternalId()
                ).id
            if affiliation_data.addresses and (address := affiliation_data.addresses[0]):
                affiliation.geo.country = address.country
                affiliation.geo.country_code = address.country_code
                affiliation.geo.region = address.state
                affiliation.geo.city = address.city
                affiliation.geo.latitude = address.lat
                affiliation.geo.longitude = address.lng
                author.countries.append(affiliation.geo.country)
        author.countries = list(set(author.countries))
    work.affiliations_data = None


def set_source_data(work: Work) -> None:
    source_data = next(filter(lambda x: x.id == work.source.id, work.source_data), Source())
    work.source.types = source_data.types
    if source_data.external_ids:
        work.source.issn_l = next(filter(lambda x: x.source == "issn_l", source_data.external_ids), ExternalId()).id
    if source_data.updated:
        work.source.is_in_doaj = (
            True if next(filter(lambda x: x.source == "doaj", source_data.updated), None) else False
        )
    work.source_data = None



from bson import ObjectId
from pydantic import BaseModel, Field

from domain.models.base_model import (
    Type,
    Updated,
    ExternalId,
    ExternalUrl,
    Name,
    PyObjectId,
    Ranking,
    Subject,
    Publisher,
    APC,
)


class Copyright(BaseModel):
    author_retains: bool | None = None
    url: str | None = None


class Licence(BaseModel):
    BY: bool | None
    NC: bool | None
    ND: bool | None
    SA: bool | None
    type: str | None
    url: str | None = None


class Waiver(BaseModel):
    has_waiver: bool | None = None
    url: str | None = None


class Source(BaseModel):
    id: PyObjectId = Field(alias="_id")
    abbreviations: list[str] | None = None
    addresses: list | None = None
    apc: APC | None = None
    copyright: Copyright | None = None
    external_ids: list[ExternalId] | None = None
    external_urls: list[ExternalUrl] | None = None
    keywords: list[str] | None = None
    languages: list[str] | None = None
    licenses: list[Licence] | None = None
    names: list[Name] | None = None
    open_access_start_year: int | None = None
    plagiarism_detection: bool | None = None
    publication_time_weeks: int | None = None
    publisher: Publisher | str | None = None
    ranking: list[Ranking] | None = None
    relations: list | None = None
    review_process: list[str] | None = None
    review_processes: list | None = None
    subjects: list[Subject] | None = None
    types: list[Type] | None = None
    updated: list[Updated] | None = None
    waiver: Waiver | None = None

    date_published: str | int | None = None
    affiliation_names: list[Name] | None = None

    class Config:
        json_encoders = {ObjectId: str}



from typing import Generator

from bson import ObjectId

from infrastructure.generators import work_generator
from domain.models.base_model import QueryParams
from infrastructure.repositories import base_repository, work_repository
from infrastructure.mongo import database


def get_works_by_affiliation_for_api_expert(
    affiliation_id: str,
    query_params: QueryParams,
    pipeline_params: dict | None = None,
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [
        {
            "$match": {"authors.affiliations.id": affiliation_id},
        }
    ]
    return get_works_for_api_expert(pipeline, pipeline_params, query_params)


def get_works_by_person_for_api_expert(
    person_id: str, query_params: QueryParams, pipeline_params: dict | None = None
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [{"$match": {"authors.id": ObjectId(person_id)}}]
    return get_works_for_api_expert(pipeline, pipeline_params, query_params)


def search_works_for_api_expert(query_params: QueryParams, pipeline_params: dict | None = None) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    return get_works_for_api_expert(pipeline, pipeline_params, query_params)


def get_works_for_api_expert(pipeline: list, pipeline_params: dict, query_params: QueryParams) -> Generator:
    base_repository.set_match(pipeline, pipeline_params.get("match"))
    if sort := query_params.sort:
        base_repository.set_sort(sort, pipeline)
    pipeline += [
        {
            "$lookup": {
                "from": "person",
                "localField": "authors.id",
                "foreignField": "_id",
                "as": "authors_data",
                "pipeline": [
                    {
                        "$project": {
                            "id": "$_id",
                            "sex": 1,
                            "birthplace.country": 1,
                            "birthdate": 1,
                            "first_names": 1,
                            "last_names": 1,
                            "affiliations.id": 1,
                            "affiliations.start_date": 1,
                            "affiliations.end_date": 1,
                            "affiliations.name": 1,
                            "affiliations.types": 1,
                            "ranking": 1,
                            "external_ids": 1,
                        }
                    },
                ],
            }
        },
        {
            "$lookup": {
                "from": "sources",
                "localField": "source.id",
                "foreignField": "_id",
                "as": "source_data",
                "pipeline": [
                    {
                        "$project": {
                            "names": 1,
                            "id": "$_id",
                            "types": 1,
                            "external_ids": 1,
                            "updated": 1,
                        }
                    },
                ],
            }
        },
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "authors.affiliations.id",
                "foreignField": "_id",
                "as": "affiliations_data",
                "pipeline": [{"$project": {"id": "$_id", "addresses": 1, "external_ids": 1}}],
            },
        },
    ]
    work_repository.set_product_filters(pipeline, query_params)
    base_repository.set_project(pipeline, pipeline_params.get("project"))
    base_repository.set_pagination(pipeline, query_params)
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)



from domain.models.base_model import QueryParams
from domain.parsers import person_parser
from infrastructure.repositories import person_repository


def get_person_by_id(person_id: str) -> dict:
    person = person_repository.get_person_by_id(person_id)
    data = person_parser.parse_person(person)
    return {"data": data}


def search_persons(query_params: QueryParams) -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "full_name",
            "affiliations",
            "external_ids",
            "citations_count",
            "products_count",
            "affiliations_data",
        ]
    }
    persons, total_results = person_repository.search_persons(query_params, pipeline_params)
    persons_list = []
    for person in persons:
        persons_list.append(person)
    data = person_parser.parse_search_result(persons_list)
    return {"data": data, "total_results": total_results}



from typing import Generator, Tuple

from bson import ObjectId

from infrastructure.generators import project_generator
from domain.models.base_model import QueryParams
from domain.models.project_model import Project
from infrastructure.repositories import base_repository
from infrastructure.mongo import database
from domain.exceptions.not_entity_exception import NotEntityException


def get_project_by_id(project_id: str) -> Project:
    project = database["projects"].find_one(ObjectId(project_id))
    if not project:
        raise NotEntityException(f"The project with id {project_id} does not exist.")
    return Project(**project)


def get_projects_by_affiliation(
    affiliation_id: str,
    query_params: QueryParams,
    pipeline_params: dict | None = None,
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [
        {
            "$match": {
                "authors.affiliations.id": affiliation_id,
            },
        },
    ]
    if sort := query_params.sort:
        base_repository.set_sort(sort, pipeline)
    base_repository.set_pagination(pipeline, query_params)
    base_repository.set_project(pipeline, pipeline_params.get("project"))
    cursor = database["projects"].aggregate(pipeline)
    return project_generator.get(cursor)


def get_projects_count_by_affiliation(affiliation_id: str) -> int:
    pipeline = get_projects_by_affiliation_pipeline(affiliation_id)
    pipeline += [{"$count": "total"}]
    return next(database["projects"].aggregate(pipeline), {"total": 0}).get("total", 0)


def get_projects_by_person(person_id: str, query_params: QueryParams, pipeline_params: dict | None = None) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    if sort := query_params.sort:
        base_repository.set_sort(sort, pipeline)
    base_repository.set_pagination(pipeline, query_params)
    base_repository.set_project(pipeline, pipeline_params.get("project"))
    cursor = database["projects"].aggregate(pipeline)
    return project_generator.get(cursor)


def get_projects_count_by_person(person_id: str) -> int:
    pipeline = [{"$match": {"authors.id": ObjectId(person_id)}}, {"$count": "total"}]
    return next(database["projects"].aggregate(pipeline), {"total": 0}).get("total", 0)


def search_projects(query_params: QueryParams, pipeline_params: dict | None = None) -> Tuple[Generator, int]:
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    base_repository.set_search_end_stages(pipeline, query_params, pipeline_params)
    projects = database["projects"].aggregate(pipeline)
    count_pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    count_pipeline += [
        {"$count": "total_results"},  # type: ignore
    ]
    total_results = next(database["projects"].aggregate(count_pipeline), {"total_results": 0}).get("total_results", 0)
    return project_generator.get(projects), total_results


def get_projects_by_affiliation_pipeline(affiliation_id: str) -> list:
    return [
        {
            "$match": {
                "authors.affiliations.id": affiliation_id,
            },
        },
    ]



from typing import Generator

from pydantic import BaseModel, field_validator, Field, conint, model_validator
from bson import ObjectId


class PyObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, value: str | ObjectId, other: None) -> str:
        if value == "":
            return str(value)
        if not ObjectId.is_valid(value):
            raise ValueError(f"Invalid ObjectId '{str(value)}'")
        return str(value)


class CitationsCount(BaseModel):
    source: str | None
    count: int | None
    provenance: str | None = None


class Type(BaseModel):
    type_class: str | None = None
    code: str | None = None
    level: int | None = None
    provenance: str | None = None
    source: str | None
    type: str | None


class Updated(BaseModel):
    time: int | None
    source: str | None


class Identifier(BaseModel):
    COD_RH: str | None
    COD_PRODUCTO: str | None = None


class ExternalId(BaseModel):
    id: str | int | list[str] | Identifier | None = None
    provenance: str | None = None
    source: str | None = None

    @field_validator("id", mode="after")
    @classmethod
    def id_validator(cls, value: str | int | list[str] | Identifier | None) -> str | int | list[str] | None:
        if isinstance(value, Identifier):
            if value.COD_PRODUCTO and value.COD_RH:
                value = value.COD_RH + "-" + value.COD_PRODUCTO
            else:
                value = value.COD_RH
        return value

    def __hash__(self) -> int:
        return hash(str(self.id) + str(self.source))


class ExternalUrl(BaseModel):
    url: str | int | None = None
    source: str | None = None
    provenance: str | None = None

    def __hash__(self) -> int:
        return hash(str(self.url) + str(self.source))


class Name(BaseModel):
    name: str | None
    lang: str | None
    source: str | None
    provenance: str | None = None


class Ranking(BaseModel):
    date: int | str | None = None
    from_date: int | str | None = None
    issn: str | None = None
    order: str | int | None = None
    rank: str | int | float | None = None
    source: str | None = None
    to_date: int | str | None = None


class Status(BaseModel):
    source: str | None
    status: str | None


class SubjectContent(BaseModel):
    id: PyObjectId | None = None
    external_ids: list[ExternalId] | None = None
    level: int | None = None
    name: str | None

    class Config:
        json_encoders = {ObjectId: str}


class Subject(BaseModel):
    provenance: str | None = None
    source: str | None
    subjects: list[SubjectContent] | None = None


class Address(BaseModel):
    city: str | None = None
    country: str | None = None
    country_code: str | None = None
    lat: float | str | None = None
    lng: float | str | None = None
    postcode: str | None = None
    state: str | None = None


class Publisher(BaseModel):
    id: str | None = None
    country_code: str | None = None
    name: str | float | None = None


class Paid(BaseModel):
    currency: str | None = None
    provenance: str | None = None
    source: str | None = None
    value: int | None = None
    value_usd: int | None = None


class APC(BaseModel):
    charges: int | None = None
    currency: str | None = None
    paid: Paid | None = None


class QueryParams(BaseModel):
    limit: conint(ge=1, le=250) | None = Field(default=None, alias="max")  # type: ignore
    page: conint(ge=1) | None = None  # type: ignore
    keywords: str | None = None
    plot: str | None = None
    sort: str | None = None
    product_types: str | None = None
    years: str | None = None
    status: str | None = None
    subjects: str | None = None
    countries: str | None = None
    groups_ranking: str | None = None
    authors_ranking: str | None = None

    @model_validator(mode="after")
    def validate_pagination_and_sort(self) -> "QueryParams":
        if not self.plot and not self.limit and not self.page and not self.sort:
            self.limit = 10
            self.page = 1
            self.sort = "citations_desc"
        return self


class Geography(BaseModel):
    city: str | None = None
    region: str | None = None
    country: str | None = None
    country_code: str | None = None
    latitude: float | str | None = None
    longitude: float | str | None = None


class Affiliation(BaseModel):
    id: str | None = None
    name: str | None = None
    types: list[Type] | None = None
    start_date: int | str | None = None
    end_date: int | str | None = None

    ror: str | None = None
    geo: Geography | None = Field(default_factory=Geography)
    addresses: list[Address] | None = None
    position: str | None = None
    ranking: list[Ranking] | None = None
    external_ids: list[ExternalId] | None = None

    class Config:
        json_encoders = {ObjectId: str}


class BirthPlace(BaseModel):
    city: str | None = None
    country: str | None = None
    state: str | None = None


class Author(BaseModel):
    id: PyObjectId | None = None
    affiliations: list[Affiliation] | None = Field(default_factory=list[Affiliation])
    full_name: str | None = None

    birth_country: str | None = None
    age: int | None = None
    birthdate: int | str | None = None
    birthplace: BirthPlace | None = None
    countries: list[str] | None = None
    first_names: list[str] | None = None
    last_names: list[str] | None = None
    sex: str | None = None
    external_ids: list[ExternalId] | None = None
    ranking: list[Ranking] | str | None = None

    @field_validator("external_ids")
    @classmethod
    def delete_sensitive_external_ids(cls, value: list | None) -> list | None:
        if value is None:
            return value
        return list(
            filter(
                lambda external_id: external_id.source
                not in ["Cédula de Ciudadanía", "Cédula de Extranjería", "Passport"],
                value,
            )
        )

    class Config:
        json_encoders = {ObjectId: str}


class Group(BaseModel):
    id: str | None = None
    name: str | None

    class Config:
        json_encoders = {ObjectId: str}


class Title(BaseModel):
    lang: str | None
    provenance: str | None = None
    source: str | None
    title: str | None


class ProductType(BaseModel):
    name: str | None = None
    source: str | None = None



from typing import Tuple

from flask import Blueprint, request, Response, jsonify
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from domain.services import (
    work_service,
    person_service,
    other_work_service,
    project_service,
    person_plot_service,
    csv_service,
    patent_service,
)

person_app_router = Blueprint("person_app_router", __name__)

"""
@api {get} /app/person/:person_id Get person by id
@apiName GetPersonById
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene un autor por su ID.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>", methods=["GET"])
def get_person_by_id(person_id: str) -> Response | Tuple[Response, int]:
    try:
        data = person_service.get_person_by_id(person_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/products Get person research products
@apiName GetPersonResearchProducts
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene los productos bibliográficos de un autor.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/products", methods=["GET"])
def get_person_research_products(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        if query_params.plot:
            data = person_plot_service.get_person_plot(person_id, query_params)
            return jsonify(data)
        data = work_service.get_works_by_person(person_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/products/filters Get person research products filters
@apiName GetPersonResearchProductsFilters
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene los filtros disponibles en los productos bibliográficos de un autor.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/products/filters", methods=["GET"])
def get_person_research_products_filters(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = work_service.get_works_filters_by_person(person_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/products/csv Get person research products csv
@apiName GetPersonResearchProductsCsv
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene los productos bibliográficos de un autor en formato CSV.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/products/csv", methods=["GET"])
def get_works_csv_by_person(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = csv_service.get_works_csv_by_person(person_id, query_params)
        response = Response(data, content_type="text/csv")
        response.headers["Content-Disposition"] = "attachment; filename=affiliation.csv"
        return response
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/other_works Get person research other works
@apiName GetPersonResearchOtherWorks
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene los productos de otro tipo de un autor.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/other_works", methods=["GET"])
def get_person_research_other_works(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = other_work_service.get_other_works_by_person(person_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/patents Get person research patents
@apiName GetPersonResearchPatents
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene las patentes de un autor.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/patents", methods=["GET"])
def get_person_research_patents(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = patent_service.get_patents_by_person(person_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/person/:person_id/research/projects Get person research projects
@apiName GetPersonResearchProjects
@apiGroup Person
@apiVersion 1.0.0
@apiDescription Obtiene los proyectos de un autor.

@apiParam {String} person_id ID del autor.
"""


@person_app_router.route("/<person_id>/research/projects", methods=["GET"])
def get_person_research_projects(person_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = project_service.get_projects_by_person(person_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from typing import Generator

from domain.models.base_model import QueryParams
from infrastructure.repositories import project_repository
from domain.services.base_service import (
    set_external_ids,
    set_external_urls,
    set_authors_external_ids,
    limit_authors,
    set_product_types,
    set_title_and_language,
)
from domain.parsers import project_parser


def get_project_by_id(project_id: str) -> dict:
    project = project_repository.get_project_by_id(project_id)
    set_external_ids(project)
    set_external_urls(project)
    limit_authors(project)
    set_authors_external_ids(project)
    set_title_and_language(project)
    set_product_types(project)
    data = project_parser.parse_project(project)
    return {"data": data}


def get_project_authors(project_id: str) -> dict:
    project = project_repository.get_project_by_id(project_id)
    set_authors_external_ids(project)
    return {"data": project.model_dump()["authors"]}


def search_projects(query_params: QueryParams) -> dict:
    pipeline_params = get_projects_by_entity_pipeline_params()
    projects, total_results = project_repository.search_projects(query_params, pipeline_params)
    projects_data = get_project_by_entity_data(projects)
    data = project_parser.parse_search_results(projects_data)
    return {"data": data, "total_results": total_results}


def get_projects_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_projects_by_entity_pipeline_params()
    projects = project_repository.get_projects_by_affiliation(affiliation_id, query_params, pipeline_params)
    projects_data = get_project_by_entity_data(projects)
    data = project_parser.parse_projects_by_entity(projects_data)
    total_results = project_repository.get_projects_count_by_affiliation(affiliation_id)
    return {"data": data, "total_results": total_results}


def get_projects_by_person(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_projects_by_entity_pipeline_params()
    projects = project_repository.get_projects_by_person(person_id, query_params, pipeline_params)
    projects_data = get_project_by_entity_data(projects)
    data = project_parser.parse_projects_by_entity(projects_data)
    total_results = project_repository.get_projects_count_by_person(person_id)
    return {"data": data, "total_results": total_results}


def get_project_by_entity_data(projects: Generator) -> list:
    projects_data = []
    for project in projects:
        limit_authors(project)
        set_title_and_language(project)
        set_product_types(project)
        projects_data.append(project)
    return projects_data


def get_projects_by_entity_pipeline_params() -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "author_count",
            "authors",
            "types",
            "titles",
            "subjects",
            "year_init",
            "year_end",
            "external_ids",
            "external_urls",
            "authors_data",
        ]
    }
    return pipeline_params



from pymongo import MongoClient
from pymongo.database import Database

from config import settings

database: Database = MongoClient(host=str(settings.MONGO_URI))[settings.MONGO_DATABASE]
calculations_database: Database = MongoClient(host=str(settings.MONGO_URI))[settings.MONGO_CALCULATIONS_DATABASE]



from urllib.parse import urlparse

from domain.constants.external_urls import external_urls_dict
from domain.models.base_model import Title, ProductType, ExternalUrl
from domain.models.other_work_model import OtherWork
from domain.models.patent_model import Patent
from domain.models.project_model import Project
from domain.models.work_model import Work
from infrastructure.repositories import person_repository


def set_title_and_language(workable: Work | OtherWork | Patent | Project) -> None:
    def order(title: Title) -> float:
        hierarchy = ["openalex", "scienti", "minciencias", "ranking", "scholar"]
        return hierarchy.index(title.source) if title.source in hierarchy else float("inf")

    first_title = sorted(workable.titles, key=order)[0]
    workable.language = first_title.lang
    workable.title = first_title.title


def set_product_types(workable: Work | OtherWork | Patent | Project) -> None:
    def order(product_type: ProductType) -> float:
        hierarchy = ["openalex", "scienti", "minciencias", "scholar"]
        return hierarchy.index(product_type.source) if product_type.source in hierarchy else float("inf")

    def filter_func(product_type: ProductType) -> bool:
        if product_type.source == "minciencias" and product_type.level == 0:
            return False
        elif product_type.source == "scienti" and product_type.level in [0, 1]:
            return False
        return True

    types = filter(filter_func, workable.types)
    product_types = list(map(lambda x: ProductType(name=x.type, source=x.source), types))
    workable.product_types = sorted(product_types, key=order)


def set_authors_external_ids(workable: Work | OtherWork | Patent | Project) -> None:
    if not workable.authors:
        return
    for author in workable.authors:
        if author.id:
            author.external_ids = person_repository.get_person_by_id(str(author.id)).external_ids


def limit_authors(workable: Work | OtherWork | Patent | Project, limit: int = 10) -> None:
    if not workable.authors:
        return
    if len(workable.authors) > limit:
        workable.authors = workable.authors[:limit]


def set_external_ids(workable: Work | OtherWork | Patent | Project) -> None:
    if not workable.external_ids:
        return
    new_external_ids = []
    for external_id in workable.external_ids:
        if external_id.source in ["minciencias", "scienti"]:
            new_external_ids.append(external_id)
        else:
            workable.external_urls.append(ExternalUrl(url=external_id.id, source=external_id.source))
    workable.external_ids = list(set(new_external_ids))


def set_external_urls(workable: Work | OtherWork | Patent | Project) -> None:
    if not workable.external_urls:
        return
    new_external_urls = []
    for external_url in workable.external_urls:
        url = str(external_url.url)
        if urlparse(url).scheme and urlparse(url).netloc:
            new_external_urls.append(external_url)
        else:
            if external_url.source in external_urls_dict.keys() and url != "":
                new_external_urls.append(
                    ExternalUrl(
                        url=external_urls_dict[external_url.source].format(id=url),
                        source=external_url.source,
                    )
                )
    workable.external_urls = list(set(new_external_urls))



from typing import Generator, Tuple


from domain.models.base_model import QueryParams
from domain.constants.institutions import institutions_list
from infrastructure.generators import affiliation_generator
from domain.models.affiliation_model import Affiliation
from infrastructure.repositories import base_repository
from infrastructure.mongo import database
from infrastructure.repositories.base_repository import set_project
from domain.exceptions.not_entity_exception import NotEntityException


def get_affiliation_by_id(affiliation_id: str) -> Affiliation:
    pipeline = [
        {"$match": {"_id": affiliation_id}},
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "relations.id",
                "foreignField": "_id",
                "as": "relations_data",
                "pipeline": [{"$project": {"id": "$_id", "external_urls": 1}}],
            }
        },
    ]
    try:
        affiliation_data = database["affiliations"].aggregate(pipeline).next()
    except:
        raise NotEntityException(f"The affiliation with id {affiliation_id} does not exist.")
    return Affiliation(**affiliation_data)


def get_affiliations_by_institution(institution_id: str, relation_type: str) -> Generator:
    pipeline = [
        {
            "$match": {
                "relations.id": institution_id,
                "types.type": relation_type,
            }
        }
    ]
    set_project(pipeline, ["_id", "names"])
    affiliations = database["affiliations"].aggregate(pipeline)
    return affiliation_generator.get(affiliations)


def get_departments_by_faculty(faculty_id: str) -> Generator:
    return get_affiliations_by_institution(faculty_id, "department")


def get_groups_by_faculty_or_department(affiliation_id: str) -> Generator:
    institution_id = (
        database["affiliations"]
        .aggregate(
            [
                {"$match": {"_id": affiliation_id}},
                {"$unwind": "$relations"},
                {"$match": {"relations.types.type": "education"}},
            ]
        )
        .next()
        .get("relations", {})
        .get("id", None)
    )
    pipeline = [
        {
            "$match": {
                "affiliations.id": affiliation_id,
            }
        },
        {"$unwind": "$affiliations"},
        {"$match": {"affiliations.types.type": "group"}},
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "affiliations.id",
                "foreignField": "_id",
                "as": "group",
                "pipeline": [
                    {
                        "$match": {
                            "relations.id": institution_id,
                            "types.type": "group",
                        },
                    },
                    {"$project": {"_id": 1, "names": 1}},
                ],
            }
        },
        {"$unwind": "$group"},
        {"$group": {"_id": "$group._id", "names": {"$first": "$group.names"}}},
    ]
    groups = database["person"].aggregate(pipeline)
    return affiliation_generator.get(groups)


def search_affiliations(
    affiliation_type: str,
    query_params: QueryParams,
    pipeline_params: dict | None = None,
) -> Tuple[Generator, int]:
    types = institutions_list if affiliation_type == "institution" else [affiliation_type]
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    pipeline += [
        {
            "$match": {
                "types.type": {"$in": types},
            }
        },
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "relations.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "relations_data",  # type: ignore
                "pipeline": [{"$project": {"id": "$_id", "external_urls": 1}}],  # type: ignore
            }
        },
        {"$project": {"works": 0}},  # type: ignore
    ]
    base_repository.set_search_end_stages(pipeline, query_params, pipeline_params)
    affiliations = database["affiliations"].aggregate(pipeline)
    count_pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    count_pipeline += [
        {"$match": {"types.type": {"$in": types}}},
        {"$count": "total_results"},  # type: ignore
    ]
    total_results = next(database["affiliations"].aggregate(count_pipeline), {"total_results": 0})["total_results"]
    return affiliation_generator.get(affiliations), total_results



external_urls_dict = {
    "doi": "https://doi.org/{id}",
    "lens": "https://www.lens.org/lens/scholar/article/{id}",
    "researchgate": "https://www.researchgate.net/profile/{id}",
    "orcid": "https://orcid.org/{id}",
    "scholar": "https://scholar.google.com/scholar?hl=en&as_sdt=0%2C5&q=info%3A{id}%3Ascholar.google.com&btnG=",
    "scholar citations": "https://scholar.google.com{id}",
    "scopus": "https://www.scopus.com/authid/detail.uri?authorId={id}",
    "openalex": "https://openalex.org/{id}",
    "ror": "https://ror.org/{id}",
    "wikidata": "https://www.wikidata.org/wiki/{id}",
    "isni": "https://isni.org/isni/{id}",
    "fundref": "https://search.crossref.org/search/funders?from_ui=yes&id={id}",
    "ucas": "https://www.ucas.com/explore/search/providers?query={id}",
    "linkedin": "https://www.linkedin.com/in/{id}",
    "ukprn": "https://www.ukrlp.co.uk/ukrlp/ukrlp_provider.page_pls_provDetails?x=&pv_status=VERIFIED&pv_vis_code=L&pn_p_id={id}",
    "wos": "https://publons.com/researcher/{id}",
}



import csv
import io
from typing import Tuple

from bson import ObjectId
from flask import Blueprint, request, jsonify, Response
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from infrastructure.mongo import database

apc_api_router = Blueprint("apc_api_router", __name__)


@apc_api_router.route("/search", methods=["GET"])
def apc_search() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
        pipeline += [
            {
                "$lookup": {
                    "from": "sources",  # type: ignore
                    "localField": "source.id",  # type: ignore
                    "foreignField": "_id",  # type: ignore
                    "as": "source_data",  # type: ignore
                    "pipeline": [  # type: ignore
                        {"$project": {"_id": 0, "apc": 1, "names": 1}},
                    ],
                }
            },
            {"$unwind": "$source_data"},  # type: ignore
            {
                "$project": {
                    "_id": 0,  # type: ignore
                    "work_doi": "$doi",  # type: ignore
                    "work_title": {"$first": "$titles.title"},  # type: ignore
                    "source_name": {"$first": "$source_data.names.name"},  # type: ignore
                    "source_apc_value": "$source_data.apc.charges",  # type: ignore
                    "source_apc_currency": "$source_data.apc.currency",  # type: ignore
                    "work_apc_currency": "$apc.paid.currency",  # type: ignore
                    "work_apc_value": "$apc.paid.value",  # type: ignore
                }
            },
        ]
        data = database["works"].aggregate(pipeline)
        fieldnames = [
            "work_doi",
            "work_title",
            "source_name",
            "source_apc_value",
            "source_apc_currency",
            "work_apc_value",
            "work_apc_currency",
        ]
        data = list(data)
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=fieldnames, escapechar="\\", quoting=csv.QUOTE_MINIMAL)
        writer.writeheader()
        writer.writerows(data)
        data = output.getvalue()
        response = Response(data, content_type="text/csv")
        response.headers["Content-Disposition"] = "attachment; filename=affiliation.csv"
        return response
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


@apc_api_router.route("/person/<person_id>", methods=["GET"])
def apc_person(person_id: str) -> Response | Tuple[Response, int]:
    try:
        pipeline = [
            {"$match": {"authors.id": ObjectId(person_id)}},
            {
                "$lookup": {
                    "from": "sources",  # type: ignore
                    "localField": "source.id",  # type: ignore
                    "foreignField": "_id",  # type: ignore
                    "as": "source_data",  # type: ignore
                    "pipeline": [  # type: ignore
                        {"$project": {"_id": 0, "apc": 1, "names": 1}},
                    ],  # type: ignore
                }
            },
            {"$unwind": "$source_data"},  # type: ignore
            {
                "$project": {
                    "_id": 0,
                    "work_doi": "$doi",
                    "work_title": {"$first": "$titles.title"},  # type: ignore
                    "source_name": {"$first": "$source_data.names.name"},  # type: ignore
                    "source_apc_value": "$source_data.apc.charges",
                    "source_apc_currency": "$source_data.apc.currency",
                    "work_apc_currency": "$apc.paid.currency",
                    "work_apc_value": "$apc.paid.value",
                }
            },
        ]
        data = database["works"].aggregate(pipeline)
        fieldnames = [
            "work_doi",
            "work_title",
            "source_name",
            "source_apc_value",
            "source_apc_currency",
            "work_apc_value",
            "work_apc_currency",
        ]
        data = list(data)
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=fieldnames, escapechar="\\", quoting=csv.QUOTE_MINIMAL)
        writer.writeheader()
        writer.writerows(data)
        data = output.getvalue()
        response = Response(data, content_type="text/csv")
        response.headers["Content-Disposition"] = "attachment; filename=affiliation.csv"
        return response
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


@apc_api_router.route("/affiliation/<affiliation_id>", methods=["GET"])
def apc_affiliation(affiliation_id: str) -> Response | Tuple[Response, int]:
    try:
        pipeline = [
            {"$match": {"authors.affiliations.id": affiliation_id}},
            {
                "$lookup": {
                    "from": "sources",  # type: ignore
                    "localField": "source.id",  # type: ignore
                    "foreignField": "_id",  # type: ignore
                    "as": "source_data",  # type: ignore
                    "pipeline": [  # type: ignore
                        {"$project": {"_id": 0, "apc": 1, "names": 1}},
                    ],  # type: ignore
                }
            },
            {"$unwind": "$source_data"},  # type: ignore
            {
                "$project": {
                    "_id": 0,
                    "work_doi": "$doi",
                    "work_title": {"$first": "$titles.title"},  # type: ignore
                    "source_name": {"$first": "$source_data.names.name"},  # type: ignore
                    "source_apc_value": "$source_data.apc.charges",
                    "source_apc_currency": "$source_data.apc.currency",
                    "work_apc_currency": "$apc.paid.currency",
                    "work_apc_value": "$apc.paid.value",
                }
            },
        ]
        data = database["works"].aggregate(pipeline)
        fieldnames = [
            "work_doi",
            "work_title",
            "source_name",
            "source_apc_value",
            "source_apc_currency",
            "work_apc_value",
            "work_apc_currency",
        ]
        data = list(data)
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=fieldnames, escapechar="\\", quoting=csv.QUOTE_MINIMAL)
        writer.writeheader()
        writer.writerows(data)
        data = output.getvalue()
        response = Response(data, content_type="text/csv")
        response.headers["Content-Disposition"] = "attachment; filename=affiliation.csv"
        return response
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from typing import Tuple

from flask import Blueprint, url_for, redirect, request, Response, jsonify
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from domain.services import api_expert_service

search_api_router = Blueprint("search_api_router", __name__)


@search_api_router.route("/person", methods=["GET"])
def search_persons() -> Response:
    return redirect(url_for("router.search_app_router.search_persons", **request.args))  # type: ignore


@search_api_router.route("/works", methods=["GET"])
def search_works() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = api_expert_service.search_works(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


@search_api_router.route("/affiliations", methods=["GET"])
def search_affiliations() -> Response:
    return redirect(url_for("router.search_app_router.search_affiliations", **request.args))  # type: ignore



from typing import Generator

from domain.models.base_model import QueryParams
from infrastructure.repositories import patent_repository
from domain.services.base_service import (
    set_external_ids,
    set_external_urls,
    set_authors_external_ids,
    limit_authors,
    set_product_types,
    set_title_and_language,
)
from domain.parsers import patent_parser


def get_patent_by_id(patent_id: str) -> dict:
    patent = patent_repository.get_patent_by_id(patent_id)
    set_external_ids(patent)
    set_external_urls(patent)
    limit_authors(patent)
    set_authors_external_ids(patent)
    set_title_and_language(patent)
    set_product_types(patent)
    data = patent_parser.parse_patent(patent)
    return {"data": data}


def get_patent_authors(patent_id: str) -> dict:
    patent = patent_repository.get_patent_by_id(patent_id)
    set_authors_external_ids(patent)
    return {"data": patent.model_dump()["authors"]}


def search_patents(query_params: QueryParams) -> dict:
    pipeline_params = get_patents_by_entity_pipeline_params()
    patents, total_results = patent_repository.search_patents(query_params, pipeline_params)
    patents_data = get_patent_by_entity_data(patents)
    data = patent_parser.parse_search_results(patents_data)
    return {"data": data, "total_results": total_results}


def get_patents_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_patents_by_entity_pipeline_params()
    patents = patent_repository.get_patents_by_affiliation(affiliation_id, query_params, pipeline_params)
    patents_data = get_patent_by_entity_data(patents)
    data = patent_parser.parse_patents_by_entity(patents_data)
    total_results = patent_repository.get_patents_count_by_affiliation(affiliation_id)
    return {"data": data, "total_results": total_results}


def get_patents_by_person(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_patents_by_entity_pipeline_params()
    patents = patent_repository.get_patents_by_person(person_id, query_params, pipeline_params)
    patents_data = get_patent_by_entity_data(patents)
    data = patent_parser.parse_patents_by_entity(patents_data)
    total_results = patent_repository.get_patents_count_by_person(person_id)
    return {"data": data, "total_results": total_results}


def get_patent_by_entity_data(patents: Generator) -> list:
    patents_data = []
    for patent in patents:
        limit_authors(patent)
        set_title_and_language(patent)
        set_product_types(patent)
        patents_data.append(patent)
    return patents_data


def get_patents_by_entity_pipeline_params() -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "author_count",
            "authors",
            "types",
            "titles",
            "subjects",
            "external_ids",
            "external_urls",
            "authors_data",
        ]
    }
    return pipeline_params



open_access_status_dict = {
    "gold": "Dorado",
    "closed": "Cerrado",
    "hybrid": "Híbrido",
    "diamond": "Diamante",
    "green": "Verde",
    "bronze": "Bronce",
    "no_info": "Sin información",
}



from typing import Generator, Tuple

from bson import ObjectId

from infrastructure.generators import work_generator
from domain.models.base_model import QueryParams
from domain.models.work_model import Work
from infrastructure.repositories import base_repository
from infrastructure.mongo import database
from domain.exceptions.not_entity_exception import NotEntityException


def get_work_by_id(work_id: str) -> Work:
    work = database["works"].find_one(ObjectId(work_id))
    if not work:
        raise NotEntityException(f"The work with id {work_id} does not exist.")
    return Work(**work)


def get_works_by_affiliation(
    affiliation_id: str,
    query_params: QueryParams,
    pipeline_params: dict | None = None,
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [
        {
            "$match": {
                "authors.affiliations.id": affiliation_id,
            },
        },
    ]
    set_product_filters(pipeline, query_params)
    base_repository.set_match(pipeline, pipeline_params.get("match"))
    if sort := query_params.sort:
        base_repository.set_sort(sort, pipeline)
    base_repository.set_pagination(pipeline, query_params)
    base_repository.set_project(pipeline, pipeline_params.get("project"))
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)


def get_works_with_source_by_affiliation(
    affiliation_id: str, query_params: QueryParams, pipeline_params: dict | None = None
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    source_project = pipeline_params.get("source_project", [])
    pipeline = [
        {"$match": {"authors.affiliations.id": affiliation_id}},
    ]
    set_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "sources",  # type: ignore
                "localField": "source.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "source",  # type: ignore
                "pipeline": [{"$project": {"_id": 1, **{p: 1 for p in source_project}}}],  # type: ignore
            }
        },
        {"$unwind": "$source"},  # type: ignore
    ]
    base_repository.set_match(pipeline, pipeline_params.get("match"))
    base_repository.set_project(pipeline, pipeline_params.get("work_project"))
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)


def get_works_count_by_affiliation(affiliation_id: str, query_params: QueryParams) -> int:
    pipeline = [
        {
            "$match": {
                "authors.affiliations.id": affiliation_id,
            },
        },
    ]
    set_product_filters(pipeline, query_params)
    pipeline += [{"$count": "total"}]  # type: ignore
    return next(database["works"].aggregate(pipeline), {"total": 0}).get("total", 0)


def get_works_by_person(person_id: str, query_params: QueryParams, pipeline_params: dict | None = None) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    set_product_filters(pipeline, query_params)
    base_repository.set_match(pipeline, pipeline_params.get("match"))
    if sort := query_params.sort:
        base_repository.set_sort(sort, pipeline)
    base_repository.set_pagination(pipeline, query_params)
    base_repository.set_project(pipeline, pipeline_params.get("project"))
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)


def get_works_with_source_by_person(
    person_id: str, query_params: QueryParams, pipeline_params: dict | None = None
) -> Generator:
    if pipeline_params is None:
        pipeline_params = {}
    source_project = pipeline_params.get("source_project", [])
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    set_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "sources",  # type: ignore
                "localField": "source.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "source",  # type: ignore
                "pipeline": [{"$project": {"_id": 1, **{p: 1 for p in source_project}}}],  # type: ignore
            }
        },
        {"$unwind": "$source"},  # type: ignore
    ]
    base_repository.set_match(pipeline, pipeline_params.get("match"))
    base_repository.set_project(pipeline, pipeline_params.get("work_project"))
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)


def get_works_count_by_person(person_id: str, query_params: QueryParams) -> int:
    pipeline = [{"$match": {"authors.id": ObjectId(person_id)}}]
    set_product_filters(pipeline, query_params)
    pipeline += [{"$count": "total"}]  # type: ignore
    return next(database["works"].aggregate(pipeline), {"total": 0}).get("total", 0)


def search_works(query_params: QueryParams, pipeline_params: dict | None = None) -> Tuple[Generator, int]:
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    set_product_filters(pipeline, query_params)
    base_repository.set_search_end_stages(pipeline, query_params, pipeline_params)
    works = database["works"].aggregate(pipeline)
    count_pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    set_product_filters(count_pipeline, query_params)
    count_pipeline += [
        {"$count": "total_results"},  # type: ignore
    ]
    total_results = next(database["works"].aggregate(count_pipeline), {"total_results": 0}).get("total_results", 0)
    return work_generator.get(works), total_results


def get_works_available_filters_by_person(person_id: str, query_params: QueryParams) -> dict:
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    return get_works_available_filters(pipeline, query_params)


def get_works_available_filters_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline = [
        {"$match": {"authors.affiliations.id": affiliation_id}},
    ]
    return get_works_available_filters(pipeline, query_params)


def get_search_works_available_filters(query_params: QueryParams, pipeline_params: dict | None = None) -> dict:
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    return get_works_available_filters(pipeline, query_params)


def get_works_available_filters(pipeline: list, query_params: QueryParams) -> dict:
    set_product_filters(pipeline, query_params)
    available_filters: dict = {}
    product_types_pipeline = pipeline.copy() + [
        {"$unwind": "$types"},
        {"$group": {"_id": "$types.source", "types": {"$addToSet": "$types"}}},
    ]
    product_types = database["works"].aggregate(product_types_pipeline)
    available_filters["product_types"] = list(product_types)
    years_pipeline = pipeline.copy() + [
        {"$match": {"year_published": {"$type": "number"}}},
        {"$group": {"_id": None, "min_year": {"$min": "$year_published"}, "max_year": {"$max": "$year_published"}}},
        {"$project": {"_id": 0, "min_year": 1, "max_year": 1}},
    ]
    years = database["works"].aggregate(years_pipeline)
    available_filters["years"] = next(years, {"min_year": None, "max_year": None})
    status_pipeline = pipeline.copy() + [
        {"$group": {"_id": "$open_access.open_access_status"}},
    ]
    status = database["works"].aggregate(status_pipeline)
    available_filters["status"] = list(status)
    subjects_pipeline = pipeline.copy() + [
        {"$unwind": "$subjects"},
        {"$unwind": "$subjects.subjects"},
        {"$match": {"subjects.subjects.level": {"$in": [0, 1]}}},
        {"$group": {"_id": "$subjects.source", "subjects": {"$addToSet": "$subjects.subjects"}}},
    ]
    subjects = database["works"].aggregate(subjects_pipeline)
    available_filters["subjects"] = list(next(subjects, {"subjects": []}).get("subjects"))  # type: ignore
    countries_pipeline = pipeline.copy() + [
        {"$unwind": "$authors"},
        {"$unwind": "$authors.affiliations"},
        {
            "$group": {
                "_id": "$authors.affiliations.country_code",
            }
        },
    ]
    countries = database["works"].aggregate(countries_pipeline)
    available_filters["countries"] = list(countries)
    groups_ranking_pipeline = pipeline.copy() + [
        {"$unwind": "$groups"},
        {"$group": {"_id": "$groups.ranking"}},
    ]
    groups_ranking = database["works"].aggregate(groups_ranking_pipeline)
    available_filters["groups_ranking"] = list(groups_ranking)
    authors_ranking_pipeline = pipeline.copy() + [
        {"$unwind": "$authors"},
        {"$group": {"_id": "$authors.ranking"}},
    ]
    authors_ranking = database["works"].aggregate(authors_ranking_pipeline)
    available_filters["authors_ranking"] = list(authors_ranking)
    return available_filters


def set_product_filters(pipeline: list, query_params: QueryParams) -> None:
    set_product_type_filters(pipeline, query_params.product_types)
    set_year_filters(pipeline, query_params.years)
    set_status_filters(pipeline, query_params.status)
    set_subject_filters(pipeline, query_params.subjects)
    set_country_filters(pipeline, query_params.countries)
    set_groups_ranking_filters(pipeline, query_params.groups_ranking)
    set_authors_ranking_filters(pipeline, query_params.authors_ranking)


def set_product_type_filters(pipeline: list, type_filters: str | None) -> None:
    if not type_filters:
        return
    match_filters = []
    for type_filter in type_filters.split(","):
        params = type_filter.split("_")
        if len(params) == 1:
            match_filters.append({"types": {"$elemMatch": {"source": params[0]}}})
        elif len(params) == 2:
            match_filters.append({"types": {"$elemMatch": {"source": params[0], "type": params[1]}}})
        elif len(params) == 3 and params[0] == "scienti":
            match_filters.append(
                {"types": {"$elemMatch": {"source": params[0], "type": params[1], "code": {"$regex": "^" + params[2]}}}}  # type: ignore
            )
    pipeline += [{"$match": {"$or": match_filters}}]


def set_year_filters(pipeline: list, years: str | None) -> None:
    if not years:
        return
    year_list = [int(year) for year in years.split(",")]
    first_year = min(year_list)
    last_year = max(year_list)
    pipeline += [{"$match": {"year_published": {"$gte": first_year, "$lte": last_year}}}]


def set_status_filters(pipeline: list, status: str | None) -> None:
    if not status:
        return
    match_filters = []
    for single_status in status.split(","):
        if single_status == "unknown":
            match_filters.append({"open_access.open_access_status": None})
        elif single_status == "open":
            match_filters.append({"open_access.open_access_status": {"$nin": [None, "closed"]}})  # type: ignore
        else:
            match_filters.append({"open_access.open_access_status": single_status})  # type: ignore
    pipeline += [{"$match": {"$or": match_filters}}]


def set_subject_filters(pipeline: list, subjects: str | None) -> None:
    if not subjects:
        return
    match_filters = []
    for subject in subjects.split(","):
        params = subject.split("_")
        if len(params) == 1:
            return
        match_filters.append({"subjects.subjects": {"$elemMatch": {"level": int(params[0]), "name": params[1]}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_country_filters(pipeline: list, countries: str | None) -> None:
    if not countries:
        return
    match_filters = []
    for country in countries.split(","):
        match_filters.append({"authors.affiliations": {"$elemMatch": {"country_code": country}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_groups_ranking_filters(pipeline: list, groups_ranking: str | None) -> None:
    if not groups_ranking:
        return
    match_filters = []
    for ranking in groups_ranking.split(","):
        match_filters.append({"groups": {"$elemMatch": {"ranking": ranking}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_authors_ranking_filters(pipeline: list, authors_ranking: str | None) -> None:
    if not authors_ranking:
        return
    match_filters = []
    for ranking in authors_ranking.split(","):
        match_filters.append({"authors": {"$elemMatch": {"ranking": ranking}}})
    pipeline += [{"$match": {"$or": match_filters}}]



from pydantic import BaseModel, Field
from domain.models.base_model import PyObjectId


class Node(BaseModel):
    degree: int = Field(default_factory=int)
    id: str | PyObjectId = Field(default_factory=str)
    label: str = Field(default_factory=str)
    size: float = Field(default_factory=float)


class Edge(BaseModel):
    coauthorships: int = Field(default_factory=int)
    size: int | float = Field(default_factory=int)
    source: str = Field(default_factory=str)
    target: str = Field(default_factory=str)


class CoauthorshipNetwork(BaseModel):
    nodes: list[Node] | None = Field(default_factory=list)
    edges: list[Edge] | None = Field(default_factory=list)


class TopWord(BaseModel):
    name: str = Field(default_factory=str)
    value: int = Field(default_factory=int)


class Calculations(BaseModel):
    id: str | PyObjectId = Field(default_factory=str, alias="_id")
    coauthorship_network: CoauthorshipNetwork = Field(default_factory=CoauthorshipNetwork)
    top_words: list[TopWord] = Field(default_factory=list)



import csv
import io

from domain.constants import countries_iso
from domain.constants.open_access_status import open_access_status_dict
from domain.constants.product_types import source_titles
from domain.models.work_model import Work


def parse_csv(works: list) -> str:
    include = [
        "title",
        "language",
        "abstract",
        "authors",
        "institutions",
        "faculties",
        "departments",
        "groups",
        "countries",
        "groups_ranking",
        "ranking",
        "issue",
        "open_access_status",
        "pages",
        "start_page",
        "end_page",
        "volume",
        "bibtex",
        "scimago_quartile",
        "openalex_citations_count",
        "scholar_citations_count",
        "subjects",
        "year_published",
        "doi",
        "publisher",
        "openalex_types",
        "scienti_types",
        "source_name",
        "source_apc",
        "source_urls",
    ]
    works_dict = [work.model_dump(include=include) for work in works]
    output = io.StringIO()
    writer = csv.DictWriter(output, fieldnames=include, escapechar="\\", quoting=csv.QUOTE_MINIMAL)
    writer.writeheader()
    writer.writerows(works_dict)
    del works_dict
    return output.getvalue()


def parse_search_results(works: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "open_access",
        "citations_count",
        "product_types",
        "year_published",
        "title",
        "subjects",
        "source",
        "external_ids",
        "external_urls",
    ]
    return [work.model_dump(include=include, exclude_none=True) for work in works]


def parse_works_by_entity(works: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "open_access",
        "citations_count",
        "product_types",
        "year_published",
        "title",
        "subjects",
        "source",
        "external_ids",
    ]
    return [work.model_dump(include=include, exclude_none=True) for work in works]


def parse_work(work: Work) -> dict:
    return work.model_dump(exclude_none=True)


def parse_api_expert(works: list) -> list:
    return [work.model_dump(exclude_none=True) for work in works]


def parse_available_filters(filters: dict) -> dict:
    available_filters: dict = {}
    if product_types := filters.get("product_types"):
        available_filters["product_types"] = parse_product_type_filter(product_types)
    if years := filters.get("years"):
        available_filters["years"] = years
    if status := filters.get("status"):
        available_filters["status"] = parse_status_filter(status)
    if subjects := filters.get("subjects"):
        available_filters["subjects"] = parse_subject_filter(subjects)
    if countries := filters.get("countries"):
        available_filters["countries"] = parse_country_filter(countries)
    if groups_ranking := filters.get("groups_ranking"):
        available_filters["groups_ranking"] = parse_groups_ranking_filter(groups_ranking)
    if authors_ranking := filters.get("authors_ranking"):
        available_filters["authors_ranking"] = parse_authors_ranking_filter(authors_ranking)
    return available_filters


def parse_authors_ranking_filter(authors_ranking: list) -> list:
    parsed_authors_ranking = []
    for ranking in authors_ranking:
        if ranking.get("_id"):
            parsed_authors_ranking.append({"value": ranking.get("_id"), "label": ranking.get("_id")})
    parsed_authors_ranking.sort(key=lambda x: x.get("label"))  # type: ignore
    return parsed_authors_ranking


def parse_groups_ranking_filter(groups_ranking: list) -> list:
    parsed_groups_ranking = []
    for ranking in groups_ranking:
        if ranking.get("_id"):
            parsed_groups_ranking.append({"value": ranking.get("_id"), "label": ranking.get("_id")})
    parsed_groups_ranking.sort(key=lambda x: x.get("label"))  # type: ignore
    return parsed_groups_ranking


def parse_country_filter(countries: list) -> list:
    parsed_countries = []
    for country in countries:
        if country.get("_id"):
            parsed_countries.append(
                {"value": country.get("_id"), "label": countries_iso.countries_dict.get(country.get("_id"), "Sin País")}
            )
    parsed_countries.sort(key=lambda x: x.get("label"))  # type: ignore
    return parsed_countries


def parse_subject_filter(subjects: list) -> list:
    parsed_subjects = []
    first_level_children = []
    second_level_children = []
    for subject in subjects:
        if subject.get("level") == 0:
            first_level_children.append({"value": "0_" + subject.get("name"), "title": subject.get("name")})
        elif subject.get("level") == 1:
            second_level_children.append({"value": "1_" + subject.get("name"), "title": subject.get("name")})
    if len(first_level_children) > 0:
        first_level_children.sort(key=lambda x: x.get("title"))  # type: ignore
        parsed_subjects.append({"value": "0", "title": "Gran área de conocimiento", "children": first_level_children})  # type: ignore
    if len(second_level_children) > 0:
        second_level_children.sort(key=lambda x: x.get("title"))  # type: ignore
        parsed_subjects.append({"value": "1", "title": "Áreas de especialidad", "children": second_level_children})  # type: ignore
    return parsed_subjects


def parse_status_filter(status: list) -> list:
    statuses = []
    open_children = []
    for oa_status in status:
        if not oa_status.get("_id"):
            statuses.append({"value": "unknown", "title": "Sin información"})
        elif oa_status.get("_id") != "closed":
            open_children.append(
                {"value": oa_status.get("_id"), "title": open_access_status_dict.get(oa_status.get("_id"))}
            )
        else:
            statuses.append({"value": "closed", "title": "Cerrado"})
    if len(open_children) > 0:
        open_children.sort(key=lambda x: x.get("title"))  # type: ignore
        statuses.append({"value": "open", "title": "Abierto", "children": open_children})  # type: ignore
    statuses.sort(key=lambda x: x.get("title"))  # type: ignore
    return statuses


def parse_product_type_filter(product_types: list) -> list:
    types = []
    for product_type in product_types:
        children = []
        if product_type.get("_id") == "crossref":
            continue

        if product_type.get("_id") == "ciarp":
            continue

        elif product_type.get("_id") == "minciencias":
            inner_types = list({element["type"]: element for element in product_type.get("types")}.values())
            for inner_type in inner_types:
                if inner_type.get("level") == 0:
                    continue
                children.append(
                    {
                        "value": product_type.get("_id") + "_" + inner_type.get("type"),
                        "title": inner_type.get("type"),
                    }
                )
        elif product_type.get("_id") == "scienti":
            second_level_children = []
            third_level_children = []
            for inner_type in product_type.get("types"):
                if inner_type.get("level") == 0:
                    children.append(
                        {
                            "value": "scienti_" + inner_type.get("type") + "_" + inner_type.get("code"),
                            "title": inner_type.get("code") + " " + inner_type.get("type"),
                            "code": inner_type.get("code"),
                            "children": [],
                        }
                    )
                elif inner_type.get("level") == 1:
                    second_level_children.append(
                        {
                            "value": "scienti_" + inner_type.get("type") + "_" + inner_type.get("code"),
                            "title": inner_type.get("code") + " " + inner_type.get("type"),
                            "code": inner_type.get("code"),
                            "children": [],
                        }
                    )
                elif inner_type.get("level") == 2:
                    third_level_children.append(
                        {
                            "value": "scienti_" + inner_type.get("type") + "_" + inner_type.get("code"),
                            "title": inner_type.get("code") + " " + inner_type.get("type"),
                            "code": inner_type.get("code"),
                        }
                    )
            second_level_children.sort(key=lambda x: x.get("title"))  # type: ignore
            third_level_children.sort(key=lambda x: x.get("title"))  # type: ignore
            for child in second_level_children:
                child["children"] = list(
                    filter(lambda x: str(x.get("code")).startswith(str(child.get("code"))), third_level_children)
                )
            for child in children:
                child["children"] = list(
                    filter(lambda x: str(x.get("code")).startswith(str(child.get("code"))), second_level_children)
                )
        else:
            for inner_type in product_type.get("types"):
                children.append(
                    {
                        "value": product_type.get("_id") + "_" + inner_type.get("type"),
                        "title": inner_type.get("type"),
                    }
                )
        children.sort(key=lambda x: x.get("title"))  # type: ignore
        types.append(
            {
                "value": product_type.get("_id"),
                "title": source_titles.get(product_type.get("_id")),
                "children": children,
            }
        )
    return types



from typing import Generator

from domain.models.base_model import QueryParams
from domain.models.work_model import Work, Abstract
from infrastructure.repositories import work_repository
from domain.services import source_service
from domain.services.base_service import (
    limit_authors,
    set_title_and_language,
    set_product_types,
    set_authors_external_ids,
    set_external_urls,
    set_external_ids,
)
from domain.parsers import work_parser


def get_work_by_id(work_id: str) -> dict:
    work = work_repository.get_work_by_id(work_id)
    set_abstract(work)
    set_external_ids(work)
    set_external_urls(work)
    limit_authors(work)
    set_authors_external_ids(work)
    set_bibliographic_info(work)
    source_service.update_work_source(work)
    set_title_and_language(work)
    set_product_types(work)
    data = work_parser.parse_work(work)
    return {"data": data}


def set_abstract(work: Work) -> None:
    def order(abstract: Abstract) -> float:
        hierarchy = ["openalex", "scholar", "scienti"]
        return hierarchy.index(abstract.source) if abstract.source in hierarchy else float("inf")

    if not work.abstracts:
        return
    first_abstract = sorted(work.abstracts, key=order)[0]
    work.abstract = first_abstract.abstract


def get_work_authors(work_id: str) -> dict:
    work = work_repository.get_work_by_id(work_id)
    set_authors_external_ids(work)
    return {"data": work.model_dump()["authors"]}


def search_works(query_params: QueryParams) -> dict:
    pipeline_params = get_works_by_entity_pipeline_params()
    works, total_results = work_repository.search_works(query_params, pipeline_params)
    works_data = get_work_by_entity_data(works)
    data = work_parser.parse_search_results(works_data)
    return {"data": data, "total_results": total_results}


def get_search_works_available_filters(query_params: QueryParams) -> dict:
    available_filters = work_repository.get_search_works_available_filters(query_params)
    return work_parser.parse_available_filters(available_filters)


def get_works_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_works_by_entity_pipeline_params()
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    works_data = get_work_by_entity_data(works)
    data = work_parser.parse_works_by_entity(works_data)
    total_results = work_repository.get_works_count_by_affiliation(affiliation_id, query_params)
    return {"data": data, "total_results": total_results}


def get_works_filters_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    available_filters = work_repository.get_works_available_filters_by_affiliation(affiliation_id, query_params)
    return work_parser.parse_available_filters(available_filters)


def get_works_by_person(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_works_by_entity_pipeline_params()
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    works_data = get_work_by_entity_data(works)
    data = work_parser.parse_works_by_entity(works_data)
    total_results = work_repository.get_works_count_by_person(person_id, query_params)
    return {"data": data, "total_results": total_results}


def get_works_filters_by_person(person_id: str, query_params: QueryParams) -> dict:
    available_filters = work_repository.get_works_available_filters_by_person(person_id, query_params)
    return work_parser.parse_available_filters(available_filters)


def get_work_by_entity_data(works: Generator) -> list:
    works_data = []
    for work in works:
        limit_authors(work)
        set_title_and_language(work)
        set_product_types(work)
        set_bibliographic_info(work)
        works_data.append(work)
    return works_data


def get_works_by_entity_pipeline_params() -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "author_count",
            "open_access",
            "authors",
            "citations_count",
            "bibliographic_info",
            "types",
            "source",
            "titles",
            "subjects",
            "year_published",
            "external_ids",
            "external_urls",
            "authors_data",
        ]
    }
    return pipeline_params


def set_bibliographic_info(work: Work) -> None:
    if not work.bibliographic_info:
        return
    work.issue = work.bibliographic_info.issue
    work.volume = work.bibliographic_info.volume
    work.bibliographic_info = None
